/*
 * File...... opticspy-tx.c
 * Author.... Joe Grand, Grand Idea Studio, Inc. [@joegrand]
 * Updated... 24 May 2018
 *
 * This program uses the MacBook Pro's keyboard backlight as a proof-of-concept 
 * optical covert channel. Optical covert channels transmit data by modulating 
 * visible light in a way that is undetectable to the human eye. However, OS X
 * modulates the LEDs at 100Hz, 75% duty cycle, which prevents the use of a 
 * high data rate and presents a noticeable flicker when data is being sent.
 * ASCII printable data is sent via the backlight's LEDs with NRZ 
 * (Non-Return-to-Zero) modulation, emulating a standard UART interface.
 *
 * More details about this experiment can be found at:
 * http://www.grandideastudio.com/portfolio/optical-covert-channels/
 *
 * Code based on "Bit Bang GPIO UART from Linux Kernel," http://www.ctrlinux.com/blog/?p=154
 * and Nick Sweeting's https://github.com/pirate/mac-keyboard-brightness (released under MIT
 * License listed at end of file)
 *
 * Developed and tested on a MacBook Pro (Retina, 15-inch, Mid 2014) w/ OS X Yosemite 
 * 10.10.5 (14F2511)
 *
 * $ clang -o opticspy-tx opticspy-tx.c -framework IOKit -framework ApplicationServices
 *
 */

#include <unistd.h>
#include <IOKit/IOKitLib.h>
#include <CoreFoundation/CoreFoundation.h>

/* Constants */
#define BIT_TIME_US     30000     // Bit time in uS
#define LOOP_DELAY      5000      // Delay in uS between transmission

char const msg_covert[] = {"NGGYU"};

/* Function prototypes */
void bitbang_string(const char*);
void bitbang_char(char);

io_connect_t getDataPort(void);
void setKeyboardBrightness(float);

/* Globals */
static io_connect_t dataPort = 0;

/* Core functions */
int main (void) 
{
    printf("Welcome to OpticSpy!\n\n");
    printf("Exfiltrating data via LED...\n");
    while(1)
    {
        bitbang_string(msg_covert); // Send pre-defined message to LED
        usleep(LOOP_DELAY);
    }
}

void bitbang_string(const char* string)
{
	int i;
 
    for (i = 0; i < strlen(string); ++i)
    {
        putchar(string[i]);          // Display character to console
        fflush(stdout);
        bitbang_char(string[i]);
    }
}

void bitbang_char(char character)
{
    int i;
 
    // Start bit
    setKeyboardBrightness(1);
    usleep(BIT_TIME_US);

    // Data bits
    for (i = 0; i < 8; ++i) 
	{
        setKeyboardBrightness((character >> i) & 1);
        usleep(BIT_TIME_US);
    }
   
    // Stop bit
    setKeyboardBrightness(0);
    usleep(BIT_TIME_US);
}

/* mac-keyboard-brightness */

io_connect_t getDataPort(void) {
    kern_return_t     kr;
    io_service_t      serviceObject;

    if (dataPort) return dataPort;

    // Look up a registered IOService object whose class is AppleLMUController
    serviceObject = IOServiceGetMatchingService(kIOMasterPortDefault, IOServiceMatching("AppleLMUController"));

    if (!serviceObject) {
        printf("Failed to connect to AppleLMUController\n");
        return 0;
    }

    // Create a connection to the IOService object
    kr = IOServiceOpen(serviceObject, mach_task_self(), 0, &dataPort);
    IOObjectRelease(serviceObject);

    if (kr != KERN_SUCCESS) {
        printf("Failed to open IoService object\n");
        return 0;
    }
    return dataPort;
}

void setKeyboardBrightness(float in) {
    kern_return_t kr;

    uint64_t inputCount  = 2;
    uint64_t inputValues[2];
    uint64_t in_unknown = 0;
    uint64_t in_brightness = in * 0xfff;

    inputValues[0] = in_unknown;
    inputValues[1] = in_brightness;

    uint32_t outputCount = 1;
    uint64_t outputValues[1];

    uint32_t out_brightness;

    kr = IOConnectCallScalarMethod(
        getDataPort(),
        2 /*kSetLEDBrightnessID*/,
        inputValues,
        inputCount,
        outputValues,
        &outputCount
    );

    out_brightness = outputValues[0];

    if (kr != KERN_SUCCESS) {
        printf("setKeyboardBrightness() error\n");
        return;
    }
}

/*

MIT License

Copyright (c) 2017 Nick Sweeting

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

